﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Runtime.InteropServices;
using System.Threading;

namespace ConsoleApplication1
{
    class Program
    {
        static Int32 DeviceIndex = 0;
        static Int16[] ResultData = new Int16[0];
        static Int32 callBackCounter = 0;

        // For bitmasks:
        static UInt16 isOutMask = 0x0000;
        static byte[] Data = new byte[] { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 };

        // State flags:
        static void ADCallbackReport(IntPtr pBuf, UInt32 BufSize, UInt32 Flags, UInt32 Context)
        {
            Int16[] receivedData = new Int16[BufSize / 2];
            Marshal.Copy(pBuf, receivedData, 0, receivedData.Length);
            ResultData = ResultData.Concat(receivedData).ToArray();
            callBackCounter++;
            AIOUSB.DIO_Write1(DeviceIndex, 8, true);
        }

        static void Main(string[] args)
        {
            double Hz = 0.0;
            AIOUSB.ADCallback callBackFunction = new AIOUSB.ADCallback(ADCallbackReport);
          
            AIOUSB.CTR_StartOutputFreq(DeviceIndex, 0, Hz);
            
            byte[] ConfigBytes = new byte[0x15]
            { 02, 02, 02, 02, 02, 02, 02, 02, 02, 02, 02, 02, 02, 02, 02, 02, //Channel range codes
                0, //Calibration code
                0x05, //Trigger and counter
                0x31, //Start and end
                0, //Oversample
                0 }; //Start and end - extended (more than 16 chanels)
            AIOUSB.ADC_SetConfig(DeviceIndex, ConfigBytes);

            AIOUSB.ADC_SetCal(DeviceIndex, ":1TO1:");

            isOutMask = 2;
            Data[1] = 0xFF;
            AIOUSB.DIO_Configure(DeviceIndex, false, isOutMask, Data);

            Hz = 500000.0 / 3.0;

            Console.WriteLine("Press Return to start sampling 768k samples at 500kHz");
            Console.ReadLine();

            var startTime = DateTime.Now.TimeOfDay.TotalMilliseconds;
            AIOUSB.DIO_Write1(DeviceIndex, 8, false);

            Boolean IO8Reset = false; 
            AIOUSB.ADC_BulkContinuousCallbackStart(DeviceIndex, 1024 * 32, 64, 0x55AA55AA, callBackFunction);
            AIOUSB.CTR_StartOutputFreq(DeviceIndex, 0, Hz);
            do
            {
                Thread.Sleep(0);
                if ((callBackCounter > 0) && (IO8Reset == false))
                {
                    AIOUSB.DIO_Write1(DeviceIndex, 8, true);                            // --J2H comment out this line or not, and test 
                    Console.WriteLine("CallBack Counter = {0}", callBackCounter);
                    IO8Reset = true;
                }
                Console.Write("\r" + ResultData.Length);
            }
            while (ResultData.Length < ((768) * 1024 ));
            AIOUSB.ADC_BulkContinuousEnd(DeviceIndex);

            var endTime = DateTime.Now.TimeOfDay.TotalMilliseconds;
            
            Console.WriteLine("Done! Total time in milleseconds = {0:F2}.", endTime - startTime);
            Console.WriteLine("Total samplecount = {0:F3}k.", (Double)ResultData.Length / 1024.0);
            Console.WriteLine("Total number of callbacks = {0}.", callBackCounter);
            Console.WriteLine("Total overhead = {0:F2}%", (((endTime - startTime) / (768 * 1024 / 500))-1) * 100);
            
            var fileWriter = new System.IO.StreamWriter("RawData.txt", false);
            Int32 counter = 0;
            
            foreach (Int16 I in ResultData)
            {
                counter++;
                if (counter % 3 == 0)
                {
                    fileWriter.WriteLine("{0}", (UInt16)I);
                }
                else
                {
                    fileWriter.Write("{0}; ", (UInt16)I);
                }
                if (counter < 4)
                {
                    Console.WriteLine("{0}", I);
                }

            }
            fileWriter.Close();
            Hz = 0;
            AIOUSB.CTR_StartOutputFreq(DeviceIndex, 0, Hz);

            Console.WriteLine("Press Return to terminate the program...");
            Console.ReadLine();
        }
    }

    static class AIOUSB
    {
        public static string ErrorMessage { get; private set; }

        private const Int32 eepromSize = 0x100; //Size of EEPROM in a device. Should be 0x200, but seems to be only 0x100

        private static Int32[] devices = GetDevices();

        private static IntPtr[] pAcquireBuffers = new IntPtr[32];
        private static Int32[] AcquireBufferSizes = new Int32[32];

        public struct DeviceInfo
        {
            public Int32 ProductID;
            public String ProductName;
            public Int32 DIOBytes;
            public Int32 Counters;
        }

        // Internal utility functions
        private static void checkIndex(Int32 DeviceIndex)
        {
            if (!devices.Contains(DeviceIndex))
            {
                throw new Exception("Invalid device index requested.");
            }
        }

        private static void checkReturnCode(UInt32 ReturnCode)
        {
            if (ReturnCode != 0)
            {
                throw new Exception("Error " + (long)ReturnCode + ": " + ((WinError)ReturnCode).ToString());
            }
        }

        // General Board Utility Functions:
        [DllImport("AIOUSB.dll", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl, EntryPoint = "GetDevices")]
        private static extern UInt32 _GetDevices();

        /// <summary>
        /// Scan for available AccesIO devices and return an array of handles
        /// </summary
        public static Int32[] GetDevices()
        {
            var deviceList = new List<Int32>(0);
            UInt32 gottenDevices = 0;

            try
            {
                gottenDevices = _GetDevices();
            }
            catch (Exception e)
            {
                ErrorMessage = e.Message;
                gottenDevices = 0;
            }

            for (int i = 0; i < sizeof(UInt32); i++)
            {
                if ((gottenDevices & (1 << i)) != 0)
                {
                    deviceList.Add(i);
                }
            }

            devices = deviceList.ToArray();

            return devices;
        }

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl, EntryPoint = "AIOUSB_CloseDevice")]
        private static extern UInt32 _CloseDevice(UInt32 DeviceIndex);

        /// <summary>
        /// Close connection to device, freeing it up for others to use.
        /// </summary
        public static void CloseDevice(Int32 DeviceIndex)
        {
            checkIndex(DeviceIndex);

            checkReturnCode(_CloseDevice((UInt32)DeviceIndex));
        }

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl, EntryPoint = "ClearDevices")]
        private static extern UInt32 _ClearDevices();

        /// <summary>
        /// Closes handles and clears records of unplugged devices.
        /// </summary
        public static void ClearDevices()
        {
            checkReturnCode(_ClearDevices());
        }

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl, EntryPoint = "GetDeviceUniqueStr")]
        private static extern UInt32 _GetDeviceUniqueStr(UInt32 DeviceIndex, IntPtr pIIDSize, IntPtr pIID);

        /// <summary>
        /// Get the unique string for specified device.
        /// </summary
        public static String GetDeviceUniqueString(Int32 DeviceIndex)
        {
            checkIndex(DeviceIndex);

            IntPtr pIIDSize = Marshal.AllocHGlobal(sizeof(UInt32));
            IntPtr pIID = Marshal.AllocHGlobal(16384); // Enough to contain the description string
            UInt32 returnVal = (UInt32)WinError.ERROR_GEN_FAILURE;
            String resultString = String.Empty;

            try
            {
                Marshal.WriteInt32(pIIDSize, 16384);
                returnVal = _GetDeviceUniqueStr((UInt32)DeviceIndex, pIIDSize, pIID);
                if (returnVal == 0)
                {
                    Int32 size = Marshal.ReadInt32(pIIDSize);
                    resultString = Marshal.PtrToStringAnsi(pIID, Marshal.ReadInt32(pIIDSize));
                }
            }
            finally
            {
                Marshal.FreeHGlobal(pIID);
                Marshal.FreeHGlobal(pIIDSize);
            }

            checkReturnCode(returnVal);

            return resultString;
        }

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl, EntryPoint = "QueryDeviceInfo")]
        private static extern UInt32 _QueryDeviceInfo(UInt32 DeviceIndex, IntPtr pPID, IntPtr pNameSize, IntPtr pName, IntPtr pDIOBytes, IntPtr pCounters);

        /// <summary>
        /// Get product ID, product name, number of DIO bytes and number of counters for specified device.
        /// </summary
        public static DeviceInfo QueryDeviceInfo(Int32 DeviceIndex)
        {
            checkIndex(DeviceIndex);

            var result = new DeviceInfo();
            UInt32 returnVal = (UInt32)WinError.ERROR_GEN_FAILURE;
            IntPtr pPID = Marshal.AllocHGlobal(sizeof(UInt32));
            IntPtr pNameSize = Marshal.AllocHGlobal(sizeof(UInt32));
            IntPtr pName = Marshal.AllocHGlobal(2048); // Enough to contain the name string
            IntPtr pDIOBytes = Marshal.AllocHGlobal(sizeof(UInt32));
            IntPtr pCounters = Marshal.AllocHGlobal(sizeof(UInt32));

            try
            {
                Marshal.WriteInt32(pNameSize, 2048);
                returnVal = _QueryDeviceInfo((UInt32)DeviceIndex, pPID, pNameSize, pName, pDIOBytes, pCounters);
                if (returnVal == 0)
                {
                    result.ProductID = Marshal.ReadInt32(pPID);
                    result.ProductName = Marshal.PtrToStringAnsi(pName, Marshal.ReadInt32(pNameSize));
                    result.DIOBytes = Marshal.ReadInt32(pDIOBytes);
                    result.Counters = Marshal.ReadInt32(pCounters);
                }
            }
            finally
            {
                Marshal.FreeHGlobal(pCounters);
                Marshal.FreeHGlobal(pDIOBytes);
                Marshal.FreeHGlobal(pName);
                Marshal.FreeHGlobal(pNameSize);
                Marshal.FreeHGlobal(pPID);
            }

            checkReturnCode(returnVal);

            return result;
        }

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl, EntryPoint = "CustomEEPROMRead")]
        private static extern UInt32 _CustomEEPROMRead(UInt32 DeviceIndex, UInt32 StartAddress, IntPtr DataSize, IntPtr Data);

        /// <summary>
        /// Read a string from the EEPROM of specified device starting at specified address.
        /// </summary
        public static String CustomEEPROMReadString(Int32 DeviceIndex, Int32 StartAddress)
        {
            checkIndex(DeviceIndex);

            if (StartAddress >= eepromSize)
            {
                throw new Exception("Physical size of EEPROM is only " + eepromSize + " bytes. Can't read beyond that.");
            }

            IntPtr pDataSize = Marshal.AllocHGlobal(sizeof(UInt32));
            IntPtr pData = Marshal.AllocHGlobal(eepromSize - StartAddress + 1); // Enough to contain the description string
            UInt32 returnVal = (UInt32)WinError.ERROR_GEN_FAILURE;
            String resultString = String.Empty;

            try
            {
                Marshal.WriteInt32(pDataSize, eepromSize - StartAddress);
                Marshal.WriteByte(pData, eepromSize - StartAddress + 1, 0); // Make sure of null-termiation
                returnVal = _CustomEEPROMRead((UInt32)DeviceIndex, (UInt32)StartAddress, pDataSize, pData);
                if (returnVal == 0)
                {
                    resultString = Marshal.PtrToStringAnsi(pData);
                }
            }
            finally
            {
                Marshal.FreeHGlobal(pData);
                Marshal.FreeHGlobal(pDataSize);
            }

            checkReturnCode(returnVal);

            return resultString;
        }

        /// <summary>
        /// Read a specified number of bytes from the EEPROM of specified device starting at specified address.
        /// </summary
        public static Byte[] CustomEEPROMReadBytes(Int32 DeviceIndex, Int32 StartAddress, Int32 NumBytes)
        {
            checkIndex(DeviceIndex);

            if (NumBytes < 1)
            {
                throw new Exception("Can't read less than one byte.");
            }

            if (StartAddress + NumBytes > eepromSize)
            {
                throw new Exception("Physical size of EEPROM is only " + eepromSize + " bytes. Can't read beyond that.");
            }

            IntPtr pDataSize = Marshal.AllocHGlobal(sizeof(UInt32));
            IntPtr pData = Marshal.AllocHGlobal(NumBytes); // Enough to contain the description string
            UInt32 returnVal = (UInt32)WinError.ERROR_GEN_FAILURE;
            var resultArray = new Byte[NumBytes];

            try
            {
                Marshal.WriteInt32(pDataSize, NumBytes);
                returnVal = _CustomEEPROMRead((UInt32)DeviceIndex, (UInt32)StartAddress, pDataSize, pData);
                if (returnVal == 0)
                {
                    for (int i = 0; i < NumBytes; i++)
                    {
                        resultArray[i] = Marshal.ReadByte(pData, i);
                    }
                }
            }
            finally
            {
                Marshal.FreeHGlobal(pData);
                Marshal.FreeHGlobal(pDataSize);
            }

            checkReturnCode(returnVal);

            return resultArray;
        }

        /// <summary>
        /// Read a single byte from the EEPROM of specified device at specified address.
        /// </summary
        public static Byte CustomEEPROMReadByte(Int32 DeviceIndex, Int32 Address)
        {
            return CustomEEPROMReadBytes(DeviceIndex, Address, 1)[0];
        }

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl, EntryPoint = "CustomEEPROMWrite")]
        private static extern UInt32 _CustomEEPROMWrite(UInt32 DeviceIndex, UInt32 StartAddress, UInt32 DataSize, IntPtr Data);

        /// <summary>
        /// Write a string to the EEPROM of specified device starting at specified address.
        /// </summary
        public static void CustomEEPROMWriteString(Int32 DeviceIndex, Int32 StartAddress, String Data)
        {
            checkIndex(DeviceIndex);

            UInt32 DataSize = (UInt32)Data.Length + 1; // The +1 is the trailing null termination char.
            IntPtr pData = Marshal.StringToHGlobalAnsi(Data);
            UInt32 returnVal = (UInt32)WinError.ERROR_GEN_FAILURE;

            try
            {
                returnVal = _CustomEEPROMWrite((UInt32)DeviceIndex, (UInt32)StartAddress, DataSize, pData);
            }
            finally
            {
                Marshal.FreeHGlobal(pData);
            }

            checkReturnCode(returnVal);
        }

        /// <summary>
        /// Write an array of bytes to the EEPROM of specified device starting at specified address.
        /// </summary
        public static void CustomEEPROMWriteBytes(Int32 DeviceIndex, Int32 StartAddress, Byte[] Data)
        {
            checkIndex(DeviceIndex);

            if (Data.Count() < 1)
            {
                throw new Exception("Can't write less than one byte.");
            }

            if (StartAddress + Data.Count() > eepromSize)
            {
                throw new Exception("Physical size of EEPROM is only " + eepromSize + " bytes. Can't write beyond that.");
            }

            UInt32 DataSize = (UInt32)Data.Count();
            IntPtr pData = Marshal.AllocHGlobal(Data.Count());
            UInt32 returnVal = (UInt32)WinError.ERROR_GEN_FAILURE;

            try
            {
                for (int i = 0; i < DataSize; i++)
                {
                    Marshal.WriteByte(pData, i, Data[i]);
                }
                returnVal = _CustomEEPROMWrite((UInt32)DeviceIndex, (UInt32)StartAddress, DataSize, pData);
            }
            finally
            {
                Marshal.FreeHGlobal(pData);
            }

            checkReturnCode(returnVal);
        }

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl, EntryPoint = "GetDeviceSerialNumber")]
        private static extern UInt32 _GetDeviceSerialNumber(UInt32 DeviceIndex, IntPtr pSerialNumber);

        /// <summary>
        /// Read the serial number of specified device.
        /// This is an internal firmaware only serial number and it is not related to the board label.
        /// </summary
        public static String GetDeviceSerialNumber(Int32 DeviceIndex)
        {
            checkIndex(DeviceIndex);

            IntPtr pSerialNumber = Marshal.AllocHGlobal(sizeof(UInt64));
            UInt32 returnVal = (UInt32)WinError.ERROR_GEN_FAILURE;
            String resultString = String.Empty;

            try
            {
                returnVal = _GetDeviceSerialNumber((UInt32)DeviceIndex, pSerialNumber);
                if (returnVal == 0)
                {
                    var resultInt = (UInt64)Marshal.ReadInt64(pSerialNumber);
                    resultString = ((UInt64)Marshal.ReadInt64(pSerialNumber)).ToString();
                }
            }
            finally
            {
                Marshal.FreeHGlobal(pSerialNumber);
            }

            checkReturnCode(returnVal);

            return resultString;
        }

        /// <summary>
        /// Scan for available AccesIO devices and return an array of handles for the ones with specified value in the first byte of EEPROM.
        /// </summary
        public static Int32[] GetDevicesByEEPROMByte(Byte Data)
        {
            var deviceList = new List<Int32>(0);

            foreach (Int32 i in GetDevices())
            {
                if (CustomEEPROMReadByte(i, 0) == Data)
                {
                    deviceList.Add(i);
                }

                CloseDevice(i);
            }

            return deviceList.ToArray();
        }

        /// <summary>
        /// Scan for available AccesIO devices and return an array of handles for the ones with specified values at specified offset.
        /// </summary
        public static Int32[] GetDevicesByEEPROMData(Int32 StartAddress, Byte[] Data)
        {
            var deviceList = new List<Int32>(0);

            foreach (Int32 i in GetDevices())
            {
                if (Enumerable.SequenceEqual(CustomEEPROMReadBytes(i, StartAddress, Data.Count()), Data))
                {
                    deviceList.Add(i);
                }

                CloseDevice(i);
            }

            return deviceList.ToArray();
        }

        /// <summary>
        /// Scan for available AccesIO devices and return an array of handles for the ones with specified string at specified offset.
        /// </summary
        public static Int32[] GetDevicesByEEPROMString(Int32 StartAddress, String Data)
        {
            var deviceList = new List<Int32>(0);

            foreach (Int32 i in GetDevices())
            {
                if (CustomEEPROMReadString(i, StartAddress) == Data)
                {
                    deviceList.Add(i);
                }

                CloseDevice(i);
            }

            return deviceList.ToArray();
        }

        /// <summary>
        /// Scan for available AccesIO devices and return the one specified serial number.
        /// This is an internal firmaware only serial number and it is not related to the board label.
        /// </summary
        public static Int32 GetDeviceBySerialNumber(String SerialNumber)
        {
            foreach (Int32 i in GetDevices())
            {
                if (GetDeviceSerialNumber(i) == SerialNumber)
                {
                    CloseDevice(i);
                    return i; // It is assumed, there is only one device with specified serial number
                }
                CloseDevice(i);
            }
            throw new Exception("Device with specified serial number not found.");
        }

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl, EntryPoint = "DIO_Configure")]
        private static extern UInt32 _DIO_Configure(UInt32 DeviceIndex, Byte bTristate, IntPtr pOutMask, IntPtr pData);

        public static void DIO_Configure(Int32 DeviceIndex, Boolean TristateAll, Int32 isOutMask, Byte[] Data)
        {
            checkIndex(DeviceIndex);

            IntPtr pIsOutMask = Marshal.AllocHGlobal(sizeof(Int32));
            IntPtr pData = Marshal.AllocHGlobal(Data.Length);
            Byte TristateAllByte = (Byte)(TristateAll ? 0xFF : 0);
            UInt32 returnVal = (UInt32)WinError.ERROR_GEN_FAILURE;

            try
            {
                Marshal.WriteInt32(pIsOutMask, isOutMask);
                Marshal.Copy(Data, 0, pData, Data.Length);
                returnVal = _DIO_Configure((UInt32)DeviceIndex, TristateAllByte, pIsOutMask, pData);
            }
            finally
            {
                Marshal.FreeHGlobal(pData);
                Marshal.FreeHGlobal(pIsOutMask);
            }

            checkReturnCode(returnVal);
        }

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl, EntryPoint = "DIO_WriteAll")]
        private static extern UInt32 _DIO_WriteAll(UInt32 DeviceIndex, IntPtr pData);
        
        /// <summary>
        /// Set all outputs on (1) or off (0). 
        /// </summary
        public static void DIO_WriteAll(Int32 DeviceIndex, Int32 Data)
        {
            checkIndex(DeviceIndex);

            IntPtr pData = Marshal.AllocHGlobal(sizeof(UInt32));
            UInt32 returnVal = (UInt32)WinError.ERROR_GEN_FAILURE;

            try
            {
                Marshal.WriteInt32(pData, Data);
                returnVal = _DIO_WriteAll((UInt32)DeviceIndex, pData);
            }
            finally
            {
                Marshal.FreeHGlobal(pData);
            }

            checkReturnCode(returnVal);
        }

        /// <summary>
        /// Set all outputs on (1) or off (0).
        /// If Data array is less than 4 Bytes, zeros are appended.
        /// </summary
        public static void DIO_WriteAll(Int32 DeviceIndex, Byte[] Data)
        {
            int i = 0;
            UInt32 DataInt = 0;

            foreach (Byte DataByte in Data)
            {
                DataInt = DataInt | (((UInt32)DataByte) << (i * 8));
                i++;
            }
            DIO_WriteAll(DeviceIndex, (Int32)DataInt);
        }

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl, EntryPoint = "DIO_Write8")]
        private static extern UInt32 _DIO_Write8(UInt32 DeviceIndex, UInt32 ByteIndex, Byte Data);

        /// <summary>
        /// Set a group of 8 outputs on (1) or off (0)
        /// </summary
        public static void DIO_Write8(Int32 DeviceIndex, Int32 ByteIndex, Byte Data)
        {
            checkIndex(DeviceIndex);

            UInt32 returnVal = (UInt32)WinError.ERROR_GEN_FAILURE;

            try
            {
                returnVal = _DIO_Write8((UInt32)DeviceIndex, (UInt32)ByteIndex, Data);
            }
            catch { }

            checkReturnCode(returnVal);
        }

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl, EntryPoint = "DIO_Write1")]
        private static extern UInt32 _DIO_Write1(UInt32 DeviceIndex, UInt32 BitIndex, Byte Data);

        /// <summary>
        /// Set a single output on (true) or off (false)
        /// </summary
        public static void DIO_Write1(Int32 DeviceIndex, Int32 BitIndex, Boolean Data)
        {
            checkIndex(DeviceIndex);

            UInt32 returnVal = (UInt32)WinError.ERROR_GEN_FAILURE;

            try
            {
                returnVal = _DIO_Write1((UInt32)DeviceIndex, (UInt32)BitIndex, (Byte)(Data ? 1 : 0));
            }
            catch { }

            checkReturnCode(returnVal);
        }

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl, EntryPoint = "DIO_ReadAll")]
        private static extern UInt32 _DIO_ReadAll(UInt32 DeviceIndex, IntPtr pData);

        /// <summary>
        /// Read all input bits
        /// </summary
        public static Int32 DIO_ReadAll(Int32 DeviceIndex)
        {
            checkIndex(DeviceIndex);

            UInt32 returnVal = (UInt32)WinError.ERROR_GEN_FAILURE;
            IntPtr pData = Marshal.AllocHGlobal(sizeof(UInt32));
            Int32 Data = 0;

            try
            {
                returnVal = _DIO_ReadAll((UInt32)DeviceIndex, pData);
                Data = Marshal.ReadInt32(pData);
            }
            finally
            {
                Marshal.FreeHGlobal(pData);
            }

            checkReturnCode(returnVal);

            return Data;
        }

        /// <summary>
        /// Read all input bits
        /// </summary
        public static void DIO_ReadAll(Int32 DeviceIndex, [In, Out] Byte[] Data)
        {
            if (Data.Count() < 1)
            {
                throw new Exception("Cannot read into empty array.");
            }

            if (Data.Count() > 4)
            {
                throw new Exception("Cannot read more than 32 input bits.");
            }

            var dataInt = DIO_ReadAll(DeviceIndex);

            for (Int32 i = 0; i < Data.Count(); i++)
            {
                Data[i] = (Byte)(dataInt >> (i * 8));
            }
        }

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl, EntryPoint = "DIO_Read8")]
        private static extern UInt32 _DIO_Read8(UInt32 DeviceIndex, UInt32 ByteIndex, IntPtr pData);

        /// <summary>
        /// Read 8 input bits
        /// </summary
        public static Byte DIO_Read8(Int32 DeviceIndex, Int32 ByteIndex)
        {
            checkIndex(DeviceIndex);

            UInt32 returnVal = (UInt32)WinError.ERROR_GEN_FAILURE;
            IntPtr pData = Marshal.AllocHGlobal(sizeof(Byte));
            Byte Data = 0;

            try
            {
                returnVal = _DIO_Read8((UInt32)DeviceIndex, (UInt32)ByteIndex, pData);
                Data = Marshal.ReadByte(pData);
            }
            finally
            {
                Marshal.FreeHGlobal(pData);
            }

            checkReturnCode(returnVal);

            return Data;
        }

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl, EntryPoint = "DIO_Read1")]
        private static extern UInt32 _DIO_Read1(UInt32 DeviceIndex, UInt32 BitIndex, IntPtr pData);

        /// <summary>
        /// Read a single input
        /// </summary
        public static Boolean DIO_Read1(Int32 DeviceIndex, Int32 BitIndex)
        {
            checkIndex(DeviceIndex);

            UInt32 returnVal = (UInt32)WinError.ERROR_GEN_FAILURE;
            IntPtr pData = Marshal.AllocHGlobal(sizeof(Byte));
            Byte Data = 0;

            try
            {
                returnVal = _DIO_Read1((UInt32)DeviceIndex, (UInt32)BitIndex, pData);
                Data = Marshal.ReadByte(pData);
            }
            finally
            {
                Marshal.FreeHGlobal(pData);
            }

            checkReturnCode(returnVal);
            return (Data == 0 ? false : true);
        }

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl, EntryPoint = "ADC_SetCal")]
        private static extern UInt32 _ADC_SetCal(UInt32 DeviceIndex, IntPtr CalFileName);

        public static void ADC_SetCal(Int32 DeviceIndex, String CalFileName)
        {
            checkIndex(DeviceIndex);

            UInt32 returnVal = (UInt32)WinError.ERROR_GEN_FAILURE;
            IntPtr pCalFileName = Marshal.StringToHGlobalAnsi(CalFileName);

            try
            {
                returnVal = _ADC_SetCal((UInt32)DeviceIndex, pCalFileName);
            }
            finally
            {
                Marshal.FreeHGlobal(pCalFileName);
            }

            checkReturnCode(returnVal);
        }

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl, EntryPoint = "ADC_SetConfig")]
        private static extern UInt32 _ADC_SetConfig(UInt32 DeviceIndex, IntPtr pConfigBuf, IntPtr ConfigBufSize);

        /// <summary>
        /// Configure analog inputs
        /// </summary
        public static void ADC_SetConfig(Int32 DeviceIndex, Byte[] ConfigBuf)
        {
            checkIndex(DeviceIndex);

            UInt32 returnVal = (UInt32)WinError.ERROR_GEN_FAILURE;
            IntPtr pConfigBuf = Marshal.AllocHGlobal(ConfigBuf.Length);
            IntPtr ConfigBufSize = Marshal.AllocHGlobal(sizeof(Int32));

            try
            {
                Marshal.WriteInt32(ConfigBufSize, ConfigBuf.Length);
                for (int i = 0; i < ConfigBuf.Length; i++)
                {
                    Marshal.WriteByte(pConfigBuf, i, ConfigBuf[i]);
                }
                returnVal = _ADC_SetConfig((UInt32)DeviceIndex, pConfigBuf, ConfigBufSize);
            }
            finally
            {
                Marshal.FreeHGlobal(pConfigBuf);
                Marshal.FreeHGlobal(ConfigBufSize);
            }

            checkReturnCode(returnVal);
        }

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl, EntryPoint = "ADC_BulkAcquire")]
        private static extern UInt32 _ADC_BulkAcquire(UInt32 DeviceIndex, UInt32 BufSize, IntPtr pBuf);

        /// <summary>
        /// Start bulk data acquisition from analog inputs
        /// </summary
        public static void ADC_BulkAcquire(Int32 DeviceIndex, Int32 BufSize)
        {
            checkIndex(DeviceIndex);

            UInt32 returnVal = (UInt32)WinError.ERROR_GEN_FAILURE;

            try
            {
                if (AcquireBufferSizes[DeviceIndex] != BufSize)
                {
                    if (pAcquireBuffers[DeviceIndex] != (IntPtr)null)
                    {
                        Marshal.FreeHGlobal(pAcquireBuffers[DeviceIndex]);
                    }
                    pAcquireBuffers[DeviceIndex] = Marshal.AllocHGlobal(BufSize);
                    AcquireBufferSizes[DeviceIndex] = BufSize;
                }

                returnVal = _ADC_BulkAcquire((UInt32)DeviceIndex, (UInt32)BufSize, pAcquireBuffers[DeviceIndex]);
            }
            catch { }

            checkReturnCode(returnVal);
        }

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl, EntryPoint = "ADC_BulkPoll")]
        private static extern UInt32 _ADC_BulkPoll(UInt32 DeviceIndex, IntPtr BytesLeft);

        /// <summary>
        /// Return number of bytes left to be input
        /// </summary>
        public static Int32 ADC_BulkPol(Int32 DeviceIndex)
        {
            checkIndex(DeviceIndex);

            UInt32 returnVal = (UInt32)WinError.ERROR_GEN_FAILURE;
            IntPtr pBytesLeft = Marshal.AllocHGlobal(sizeof(Int32));
            Int32 bytesLeft;

            try
            {
                returnVal = _ADC_BulkPoll((UInt32)DeviceIndex, pBytesLeft);
                bytesLeft = Marshal.ReadInt32(pBytesLeft);
            }
            finally
            {
                Marshal.FreeHGlobal(pBytesLeft);
            }

            checkReturnCode(returnVal);

            return bytesLeft;
        }

        /// <summary>
        /// Return an array of bytes that has already been
        /// </summary>
        public static Byte[] getBulkBytes(Int32 DeviceIndex)
        {
            Int32 remaining = ADC_BulkPol(DeviceIndex);

            Byte[] data = new Byte[AcquireBufferSizes[DeviceIndex] - remaining];

            Marshal.Copy(pAcquireBuffers[DeviceIndex], data, 0, data.Length);

            return data;
        }

        /// <summary>
        /// Return an array of bytes that has already been
        /// </summary>
        public static Int16[] getBulkInt16(Int32 DeviceIndex)
        {
            Int32 remaining = ADC_BulkPol(DeviceIndex);

            Int16[] data = new Int16[(AcquireBufferSizes[DeviceIndex] - remaining)/2];

            Marshal.Copy(pAcquireBuffers[DeviceIndex], data, 0, data.Length);

            return data;
        }

        [UnmanagedFunctionPointer(CallingConvention.Cdecl)]
        public delegate void ADCallback(IntPtr pBuf, uint BufSize, uint Flags, uint Context);

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl, EntryPoint = "ADC_BulkContinuousCallbackStart")]
        private static extern UInt32 _ADC_BulkContinuousCallbackStart(UInt32 DeviceIndex, UInt32 BufSize, UInt32 BaseBufCount, UInt32 Context, ADCallback pCallback);

        public static void ADC_BulkContinuousCallbackStart(Int32 DeviceIndex, Int32 BufSize, Int32 BaseBufCount, Int32 Context, ADCallback pCallback)
        {
            checkIndex(DeviceIndex);
            UInt32 returnVal = (UInt32)WinError.ERROR_GEN_FAILURE;

            try
            {
                returnVal = _ADC_BulkContinuousCallbackStart((UInt32)DeviceIndex, (UInt32)BufSize, (UInt32)BaseBufCount, (UInt32)Context, pCallback);
            }
            catch { }

            checkReturnCode(returnVal);
        }

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl, EntryPoint = "ADC_BulkContinuousCallbackStartClocked")]
        private static extern UInt32 _ADC_BulkContinuousCallbackStartClocked(UInt32 DeviceIndex, UInt32 BufSize, UInt32 BaseBufCount, UInt32 Context, ADCallback pCallback, IntPtr pHz);

        public static void ADC_BulkContinuousCallbackStartClocked(Int32 DeviceIndex, Int32 BufSize, Int32 BaseBufCount, Int32 Context, ADCallback pCallback, Double Hz)
        {
            checkIndex(DeviceIndex);
            UInt32 returnVal = (UInt32)WinError.ERROR_GEN_FAILURE;
            IntPtr pHz = Marshal.AllocHGlobal(sizeof(Double));

            try
            {
                Double[] Hz_Array = new Double[1] { Hz };
                Marshal.Copy(Hz_Array, 0, pHz, 1);

                returnVal = _ADC_BulkContinuousCallbackStartClocked((UInt32)DeviceIndex, (UInt32)BufSize, (UInt32)BaseBufCount, (UInt32)Context, pCallback, pHz);
            }
            finally
            {
                Marshal.FreeHGlobal(pHz);
            }

            checkReturnCode(returnVal);
        }

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl, EntryPoint = "ADC_BulkContinuousEnd")]
        private static extern UInt32 _ADC_BulkContinuousEnd(UInt32 DeviceIndex, IntPtr pIOStatus);

        public static int ADC_BulkContinuousEnd(Int32 DeviceIndex)
        {
            checkIndex(DeviceIndex);

            UInt32 returnVal = (UInt32)WinError.ERROR_GEN_FAILURE;
            IntPtr pIOStatus = Marshal.AllocHGlobal(sizeof(Int32));
            Int32 ioStatus;

            try
            {
                returnVal = _ADC_BulkContinuousEnd((UInt32)DeviceIndex, pIOStatus);
                ioStatus = Marshal.ReadInt32(pIOStatus);
            }
            finally
            {
                Marshal.FreeHGlobal(pIOStatus);
            }

            checkReturnCode(returnVal);

            return ioStatus;
        }

        [DllImport("AIOUSB.DLL", CharSet = CharSet.Ansi, CallingConvention = CallingConvention.Cdecl, EntryPoint = "CTR_StartOutputFreq")]
        private static extern UInt32 _CTR_StartOutputFreq(UInt32 DeviceIndex, UInt32 BlockIndex, IntPtr pHz);

        public static void CTR_StartOutputFreq(Int32 DeviceIndex, Int32 BlockIndex, Double Hz)
        {
            checkIndex(DeviceIndex);

            UInt32 returnVal = (UInt32)WinError.ERROR_GEN_FAILURE;
            IntPtr pHz = Marshal.AllocHGlobal(sizeof(Double));

            try
            {
                Double[] Hz_Array = new Double[1] {Hz};
                Marshal.Copy(Hz_Array, 0, pHz, 1);
                returnVal = _CTR_StartOutputFreq((UInt32)DeviceIndex, (UInt32)BlockIndex, pHz);
            }
            finally
            {
                Marshal.FreeHGlobal(pHz);
            }

            checkReturnCode(returnVal);
        }
    }

}
